from __future__ import annotations

import enum

LINE_END = b"\r\n"


class Commands(bytes, enum.Enum):
    CONFIG = b"config"
    GET = b"get"
    GETS = b"gets"
    GAT = b"gat"
    GATS = b"gats"
    SET = b"set"
    CAS = b"cas"
    ADD = b"add"
    REPLACE = b"replace"
    INCR = b"incr"
    DECR = b"decr"
    APPEND = b"append"
    PREPEND = b"prepend"
    TOUCH = b"touch"
    DELETE = b"delete"
    FLUSH_ALL = b"flush_all"
    VERSION = b"version"
    STATS = b"stats"


class Responses(bytes, enum.Enum):
    OK = b"OK"
    STORED = b"STORED"
    EXISTS = b"EXISTS"
    DELETED = b"DELETED"
    TOUCHED = b"TOUCHED"
    VALUE = b"VALUE"
    NOT_FOUND = b"NOT_FOUND"
    NOT_STORED = b"NOT_STORED"
    VERSION = b"VERSION"
    ERROR = b"ERROR"
    SERVER_ERROR = b"SERVER_ERROR"
    CLIENT_ERROR = b"CLIENT_ERROR"
    STAT = b"STAT"
    END = b"END"



from __future__ import annotations

import sys

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
    from typing_extensions import NotRequired, Self, Unpack
else:
    from asyncio import timeout as asyncio_timeout
    from typing import NotRequired, Self, Unpack

__all__ = ["asyncio_timeout", "NotRequired", "Self", "Unpack"]



from __future__ import annotations

from typing import Literal

ENCODING = "utf-8"

#: Minimum number of connections to retain in the pool
MIN_CONNECTIONS = 1
#: Maximum connections to grow the pool to
MAX_CONNECTIONS = 2
#: Maximum time to leave a connection idle before disconnecting
IDLE_CONNECTION_TIMEOUT = 10.0
#: Maximum time to wait to retrieve a connection from the pool
BLOCKING_TIMEOUT = 5.0

#: Maxiumum time to wait to establish a connection
CONNECT_TIMEOUT = 1.0
#: Maxiumum time to wait to read a response for a request
READ_TIMEOUT = None
#: Maxiumum number of concurrent requests to pipeline on each connection
MAX_INFLIGHT_REQUESTS_PER_CONNECTION = 100
#:  Threshold for allowing the connection to be reused when there are requests pending.
MAX_AVERAGE_RESPONSE_TIME_FOR_CONNECTION_REUSE = 0.05

#: Whether to remove unhealthy endpoints on connection errors.
#: This is the default value for :attr:`~memcachio.endpointHealthcheckConfig.remove_unhealthy_endpoints`
REMOVE_UNHEALTHY_ENDPOINTS = False
#: Maximum numbers of errors to tolerate before marking an endpoint
#: as unhealthy
MAXIMUM_ERROR_COUNT_FOR_ENDPOINT_REMOVAL = 2
#: Whether to monitor unhealthy endpoints after they have been
#: removed and attempt to restore them if they recover
#: This is the default value for :attr:`~memcachio.endpointHealthcheckConfig.monitor_unhealthy_endpoints`
MONITOR_UNHEALTHY_ENDPOINTS = False
#: Maximum attempts to make to recover unhealthy endpoints
#: This is the default value for :attr:`~memcachio.endpointHealthcheckConfig.maximum_recovery_attempts`
MAXIMUM_RECOVERY_ATTEMPTS = 4
#:
RETRY_BACKOFF_POLICY: Literal["linear"] = "linear"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from .commands import SetCommand

if TYPE_CHECKING:
    from .connection import BaseConnection


class Authenticator(ABC):
    """
    Abstract class for authentication strategy used by
    :class:`memcachio.BaseConnection`.
    """

    @abstractmethod
    async def authenticate(self, connection: BaseConnection) -> bool:
        """
        The method that will be called immediately after a connection
        is established to the server. This should perform any authentication
        handshake required by the server. See :class:`memcachio.SimpleAuthenticator`
        for an example.

        :param connection: The connection instance that was established
        """
        ...


class SimpleAuthenticator(Authenticator):
    """
    A username/password authentication strategy for the ASCII
    protocol as defined in the `memcached protocol documentation
    <https://github.com/memcached/memcached/blob/master/doc/protocol.txt#L186>`_.
    """

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    async def authenticate(self, connection: BaseConnection) -> bool:
        """
        Authenticate the connection using a fake ``set`` command with the
        username and password as the payload.
        """
        auth_command = SetCommand("auth", f"{self.username} {self.password}")
        connection.create_request(auth_command)
        return await auth_command.response


class MemCachierAuthenticator(SimpleAuthenticator):
    """
    A username/password authentication strategy for the ASCII
    protocol as defined by the `MemCachier documentation
    <https://www.memcachier.com/documentation/supported-protocols-ascii-binary>_

    To use with a memcachier instance::

        from memcachio import Client
        from memcachio.authentication import MemCachierAuthenticator

        mc_auth = MemCachierAuthenticator("your-memcachier-username", "your-memcachier-password")
        client = Client(("mcXXX.dev.ec2.memcachier.com", 11211), authenticator=mc_auth)


    """

    async def authenticate(self, connection: BaseConnection) -> bool:
        """
        Authenticate the connection using a ``set`` command with the
        username as the key and password as the payload.
        """
        auth_command = SetCommand(self.username, self.password)
        connection.create_request(auth_command)
        return await auth_command.response



from __future__ import annotations

import hashlib
from collections.abc import Callable

from .errors import NoAvailableNodes
from .types import SingleMemcachedInstanceEndpoint


def md5_hasher(key: str) -> int:
    return int(hashlib.md5(key.encode("utf-8")).hexdigest(), 16)


class KeyRouter:
    def __init__(
        self,
        nodes: set[SingleMemcachedInstanceEndpoint] | None = None,
        hasher: Callable[[str], int] | None = None,
    ) -> None:
        """
        Rendezvous Hashing implementation to route keys to memcached
        nodes.

        :param nodes: set of memcached nodes that are candidates
        :param hasher: function to use to hash a key to a node. If not
         provided, a default implementation using :func:`hashlib.md5` from
         the standard library will be used.
        """
        self._hasher = hasher or md5_hasher
        self.nodes: set[SingleMemcachedInstanceEndpoint] = nodes or set()

    def add_node(self, node: SingleMemcachedInstanceEndpoint) -> None:
        """
        Add a node to the set of candidate nodes
        """
        self.nodes.add(node)

    def remove_node(self, node: SingleMemcachedInstanceEndpoint) -> None:
        """
        Remove a node from the set of candidate nodes
        """
        self.nodes.discard(node)

    def get_node(self, key: str) -> SingleMemcachedInstanceEndpoint:
        """
        Get the node associated with ``key``
        """
        if not self.nodes:
            raise NoAvailableNodes()
        return max(self.nodes, key=lambda node: self._hasher(f"{node}:{key}"))



from __future__ import annotations

import asyncio
import dataclasses
import enum
import logging
import time
import weakref
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from contextlib import closing, suppress
from typing import Any, Literal, TypeVar, cast

from .commands import AWSAutoDiscoveryConfig, Command
from .compat import Unpack, asyncio_timeout
from .connection import (
    BaseConnection,
    ConnectionParams,
    TCPConnection,
    UnixSocketConnection,
)
from .defaults import (
    BLOCKING_TIMEOUT,
    CONNECT_TIMEOUT,
    IDLE_CONNECTION_TIMEOUT,
    MAX_CONNECTIONS,
    MAXIMUM_ERROR_COUNT_FOR_ENDPOINT_REMOVAL,
    MAXIMUM_RECOVERY_ATTEMPTS,
    MIN_CONNECTIONS,
    MONITOR_UNHEALTHY_ENDPOINTS,
    READ_TIMEOUT,
    REMOVE_UNHEALTHY_ENDPOINTS,
    RETRY_BACKOFF_POLICY,
)
from .errors import ConnectionNotAvailable, MemcachioConnectionError
from .routing import KeyRouter
from .types import (
    AWSAutoDiscoveryEndpoint,
    MemcachedEndpoint,
    SingleMemcachedInstanceEndpoint,
    UnixSocketEndpoint,
    normalize_endpoint,
    normalize_single_server_endpoint,
)

R = TypeVar("R")

logger = logging.getLogger(__name__)


class EndpointStatus(enum.IntEnum):
    """
    Enumeration of endpoint statuses.
    Used by :meth:`~ClusterPool.update_endpoint_status`
    """

    #: Mark the endpoint as up and usable
    UP = enum.auto()
    #: Mark the endpoint as down and not in use
    DOWN = enum.auto()


@dataclasses.dataclass
class EndpointHealthcheckConfig:
    #: Whether to remove unhealthy endpoints on connection errors
    remove_unhealthy_endpoints: bool = REMOVE_UNHEALTHY_ENDPOINTS
    #: Maximum numbers of errors to tolerate before marking an endpoint
    #: as unhealthy
    maximum_error_count_for_removal: int = MAXIMUM_ERROR_COUNT_FOR_ENDPOINT_REMOVAL
    #: Whether to monitor unhealthy endpoints after they have been
    #: removed and attempt to restore them if they recover
    monitor_unhealthy_endpoints: bool = MONITOR_UNHEALTHY_ENDPOINTS
    #: Maximum attempts to make to recover unhealthy endpoints
    maximum_recovery_attempts: int = MAXIMUM_RECOVERY_ATTEMPTS
    #: Retry backoff policy
    retry_backoff_policy: Literal["linear", "exponential"] = RETRY_BACKOFF_POLICY


@dataclasses.dataclass
class PoolMetrics:
    """
    Tracks metrics for a connection pool.
    """

    #: Timestamp when the pool was initialized.
    created_at: float | None = None
    #: Total number of successfully processed requests.
    requests_processed: int = 0
    #: Total number of requests that failed.
    requests_failed: int = 0
    #: Timestamp when the last connection was established
    last_connection_created: float = 0.0
    #: Timestamp when the last request completed processing.
    last_request_processed: float = 0.0
    #: Average time taken to process requests.
    average_response_time: float = 0.0
    #: Number of requests currently pending.
    requests_pending: int = 0
    #: Current connections
    current_connections: int = 0
    #: Maximum connections created
    maximum_connections: int = 0
    #: Total number of connection errors
    connection_errors: int = 0
    #: Current endpoint status
    status: EndpointStatus | None = None
    #: Number of times the pool was marked down
    down_count: int = 0
    #: Number of times the node was switched from down to up
    recovery_count: int = 0

    def on_connection_created(self, connection: BaseConnection) -> None:
        self.last_connection_created = time.time()
        self.current_connections += 1
        self.maximum_connections = max(self.maximum_connections, self.current_connections)
        if not self.status:
            self.status = EndpointStatus.UP

    def on_connection_error(self, connection: BaseConnection | None, exc: Exception) -> None:
        self.connection_errors += 1
        if not self.status:
            self.status = EndpointStatus.UP

    def on_connection_terminated(self, connection: BaseConnection) -> None:
        self.current_connections -= 1

    def on_command_dispatched(self, command: Command[Any]) -> None:
        self.requests_pending += 1
        pass

    def on_status_update(self, status: EndpointStatus) -> None:
        if self.status != status:
            match status:
                case EndpointStatus.DOWN:
                    self.down_count += 1
                case EndpointStatus.UP:
                    self.recovery_count += 1
        self.status = status

    def on_command_completed(self, command: Command[Any]) -> None:
        self.requests_pending -= 1
        self.last_request_processed = time.time()
        if not command.noreply:
            if command.response.done() and not command.response.cancelled():
                if not command.response.exception():
                    self.average_response_time = (
                        (self.requests_processed * self.average_response_time)
                        + command.response_time
                    ) / (self.requests_processed + 1)
                    self.requests_processed += 1
                else:
                    self.requests_failed += 1
        elif command.request_sent.done():
            self.requests_processed += 1

    @classmethod
    def merge(cls, metrics: Iterable[PoolMetrics]) -> PoolMetrics:
        if not metrics:
            return PoolMetrics()
        return PoolMetrics(
            created_at=min(metrics, key=lambda m: m.created_at or 0).created_at,
            requests_processed=sum([m.requests_processed for m in metrics]),
            requests_failed=sum([m.requests_failed for m in metrics]),
            requests_pending=sum([m.requests_pending for m in metrics]),
            last_connection_created=max(
                metrics, key=lambda m: m.last_connection_created
            ).last_connection_created,
            last_request_processed=max(
                metrics, key=lambda m: m.last_request_processed
            ).last_request_processed,
            average_response_time=sum(
                [m.requests_processed * m.average_response_time for m in metrics]
            )
            / (sum([m.requests_processed for m in metrics]) or 1),
            current_connections=sum([m.current_connections for m in metrics]),
            maximum_connections=sum([m.maximum_connections for m in metrics]),
            connection_errors=sum([m.connection_errors for m in metrics]),
            recovery_count=sum([m.recovery_count for m in metrics]),
            down_count=sum([m.down_count for m in metrics]),
            status=EndpointStatus.UP
            if all([m.status == EndpointStatus.UP for m in metrics])
            else EndpointStatus.DOWN,
        )


class Pool(ABC):
    """
    The abstract base class for a connection pool used by
    :class:`~memcachio.Client`
    """

    def __init__(
        self,
        endpoint: MemcachedEndpoint,
        min_connections: int = MIN_CONNECTIONS,
        max_connections: int = MAX_CONNECTIONS,
        blocking_timeout: float = BLOCKING_TIMEOUT,
        idle_connection_timeout: float = IDLE_CONNECTION_TIMEOUT,
        **connection_args: Unpack[ConnectionParams],
    ):
        """
        :param endpoint: The memcached server address(es)
        :param min_connections: The minimum number of connections to keep in the pool.
        :param max_connections: The maximum number of simultaneous connections to memcached.
        :param blocking_timeout: The timeout (in seconds) to wait for a connection to become available.
        :param idle_connection_timeout: The maximum time to allow a connection to remain idle in the pool
         before being disconnected
        :param connection_args: Arguments to pass to the constructor of :class:`~memcachio.BaseConnection`.
         refer to :class:`~memcachio.connection.ConnectionParams`
        """
        self.endpoint = normalize_endpoint(endpoint)
        self._max_connections = max_connections
        self._min_connections = min_connections
        self._blocking_timeout = blocking_timeout
        self._idle_connection_timeout = idle_connection_timeout
        self._connection_parameters: ConnectionParams = connection_args

    @property
    @abstractmethod
    def metrics(self) -> PoolMetrics:
        """
        Pool health metrics
        """
        ...

    @abstractmethod
    async def initialize(self) -> None:
        """
        Initialize the connection pool. The method can throw
        a connection error if the target server(s) can't be connected
        to.
        """
        ...

    @abstractmethod
    async def execute_command(self, command: Command[R]) -> None:
        """
        Dispatches a command to memcached. To receive the response the future
        pointed to by ``command.response`` should be awaited as it will be updated
        when the response (or exception) is available on the transport.
        """
        ...

    @abstractmethod
    def close(self) -> None:
        """
        Closes the connection pool and disconnects all active connections
        """
        ...

    def __del__(self) -> None:
        with suppress(RuntimeError):
            self.close()


class SingleServerPool(Pool):
    """
    Connection pool to manage connections to a single memcached
    server.
    """

    def __init__(
        self,
        endpoint: SingleMemcachedInstanceEndpoint,
        min_connections: int = MIN_CONNECTIONS,
        max_connections: int = MAX_CONNECTIONS,
        blocking_timeout: float = BLOCKING_TIMEOUT,
        idle_connection_timeout: float = IDLE_CONNECTION_TIMEOUT,
        **connection_args: Unpack[ConnectionParams],
    ) -> None:
        super().__init__(
            endpoint,
            min_connections=min_connections,
            max_connections=max_connections,
            blocking_timeout=blocking_timeout,
            idle_connection_timeout=idle_connection_timeout,
            **connection_args,
        )
        self.__server_endpoint = endpoint
        self.__connections: asyncio.Queue[BaseConnection | None] = asyncio.LifoQueue(
            self._max_connections
        )
        self.__pool_lock: asyncio.Lock = asyncio.Lock()
        self._connection_class: type[TCPConnection | UnixSocketConnection]
        self.__initialized = False
        self._connection_parameters.setdefault("on_connect_callbacks", []).append(
            self.__on_connection_created
        )
        self._connection_parameters.setdefault("on_disconnect_callbacks", []).append(
            self.__on_connection_disconnected
        )
        self._active_connections: weakref.WeakSet[BaseConnection] = weakref.WeakSet()
        self.__metrics = PoolMetrics()

        while True:
            try:
                self.__connections.put_nowait(None)
            except asyncio.QueueFull:
                break

    @property
    def metrics(self) -> PoolMetrics:
        return self.__metrics

    async def initialize(self) -> None:
        if self.__initialized:
            return
        async with self.__pool_lock:
            if self.__initialized:
                return
            if not self.metrics.created_at:
                self.metrics.created_at = time.time()
            for _ in range(self._min_connections):
                connection = self.__connections.get_nowait()
                try:
                    if not connection:
                        self.__connections.put_nowait(await self.__create_connection())
                except ConnectionError as e:
                    self.metrics.on_connection_error(connection, e)
                    self.__connections.put_nowait(None)
                    raise
            self.__initialized = True

    async def execute_command(self, command: Command[R]) -> None:
        connection, release = None, None
        try:
            connection, release = await self.__get_connection_from_pool()
            await connection.connect()
            connection.create_request(command)
            self.metrics.on_command_dispatched(command)
            await command.request_sent
            (command.request_sent if command.noreply else command.response).add_done_callback(
                lambda _: self.metrics.on_command_completed(command)
            )
            if release:
                if command.noreply:
                    self.__connections.put_nowait(connection)
                else:
                    command.response.add_done_callback(
                        lambda _: self.__connections.put_nowait(connection)
                    )
        except MemcachioConnectionError as e:
            self.metrics.on_connection_error(connection, e)
            if release:
                self.__connections.put_nowait(None)
            raise

    def close(self) -> None:
        while True:
            try:
                if connection := self.__connections.get_nowait():
                    connection.close()
            except asyncio.QueueEmpty:
                break
        while True:
            try:
                self.__connections.put_nowait(None)
            except asyncio.QueueFull:
                break
        self.__initialized = False

    async def __get_connection_from_pool(self) -> tuple[BaseConnection, bool]:
        await self.initialize()
        released = False
        try:
            async with asyncio_timeout(self._blocking_timeout):
                connection = await self.__connections.get()
                try:
                    if connection and connection.reusable:
                        self.__connections.put_nowait(connection)
                        released = True
                    else:
                        if not connection:
                            connection = await self.__create_connection()
                except ConnectionError as e:
                    self.metrics.on_connection_error(connection, e)
                    self.__connections.put_nowait(None)
                    raise
            return connection, not released
        except asyncio.TimeoutError:
            raise ConnectionNotAvailable(self.__server_endpoint, self._blocking_timeout)

    async def __create_connection(self) -> BaseConnection:
        connection: BaseConnection
        if isinstance(self.__server_endpoint, UnixSocketEndpoint):
            connection = UnixSocketConnection(self.__server_endpoint, **self._connection_parameters)
        else:
            connection = TCPConnection(self.__server_endpoint, **self._connection_parameters)
        if not connection.connected:
            await connection.connect()
        return connection

    def __check_connection_idle(self, connection: BaseConnection) -> None:
        if (
            time.time() - connection.metrics.last_read > self._idle_connection_timeout
            and connection.metrics.requests_pending == 0
            and len(self._active_connections) > self._min_connections
        ):
            connection.close()
            self._active_connections.discard(connection)
        elif connection.connected:
            asyncio.get_running_loop().call_later(
                self._idle_connection_timeout, self.__check_connection_idle, connection
            )

    def __on_connection_created(self, connection: BaseConnection) -> None:
        self._active_connections.add(connection)
        self.metrics.on_connection_created(connection)
        if self._idle_connection_timeout:
            asyncio.get_running_loop().call_later(
                self._idle_connection_timeout, self.__check_connection_idle, connection
            )

    def __on_connection_disconnected(self, connection: BaseConnection) -> None:
        self.metrics.on_connection_terminated(connection)
        self._active_connections.discard(connection)


class ClusterPool(Pool):
    """
    Connection pool to manage connections to multiple memcached
    servers.

    For multi-key commands, rendezvous hashing is used to distribute the command
    to the appropriate endpoints.
    """

    def __init__(
        self,
        endpoint: Sequence[SingleMemcachedInstanceEndpoint] | AWSAutoDiscoveryEndpoint,
        min_connections: int = MIN_CONNECTIONS,
        max_connections: int = MAX_CONNECTIONS,
        blocking_timeout: float = BLOCKING_TIMEOUT,
        idle_connection_timeout: float = IDLE_CONNECTION_TIMEOUT,
        hashing_function: Callable[[str], int] | None = None,
        endpoint_healthcheck_config: EndpointHealthcheckConfig | None = None,
        **connection_args: Unpack[ConnectionParams],
    ) -> None:
        """
        :param endpoint: The memcached server address(es)
        :param min_connections: The minimum number of connections per endpoint to keep in the pool.
        :param max_connections: The maximum number of simultaneous connections per  memcached endpoint.
        :param blocking_timeout: The timeout (in seconds) to wait for a connection to become available.
        :param idle_connection_timeout: The maximum time to allow a connection to remain idle in the pool
         before being disconnected
        :param hashing_function: A function to use for routing keys to
         endpoints for multi-key commands. If none is provided the default
         :func:`hashlib.md5` implementation from the standard library is used.
        :param endpoint_healthcheck_config: Configuration to control whether
         endpoints are automatically removed/recovered based on health checks.
        :param connection_args: Arguments to pass to the constructor of :class:`~memcachio.BaseConnection`.
         refer to :class:`~memcachio.connection.ConnectionParams`
        """
        self._cluster_pools: dict[SingleMemcachedInstanceEndpoint, SingleServerPool] = {}
        self.__pool_lock = asyncio.Lock()
        self.__initialized = False
        super().__init__(
            endpoint,
            min_connections=min_connections,
            max_connections=max_connections,
            blocking_timeout=blocking_timeout,
            idle_connection_timeout=idle_connection_timeout,
            **connection_args,
        )
        self.__autodiscovery_current_version: int = 0
        self.__autodiscovery_endpoint: AWSAutoDiscoveryEndpoint | None = None
        self.__all_endpoints: set[SingleMemcachedInstanceEndpoint] = set()
        if isinstance(self.endpoint, AWSAutoDiscoveryEndpoint):
            self.__autodiscovery_endpoint = self.endpoint
        else:
            self.__all_endpoints = {
                normalize_single_server_endpoint(endpoint)
                for endpoint in cast(Iterable[SingleMemcachedInstanceEndpoint], self.endpoint)
            }
        self.__unhealthy_endpoints: set[SingleMemcachedInstanceEndpoint] = set()
        self._router = KeyRouter(self.__all_endpoints, hasher=hashing_function)
        self.__healthcheck_tasks: dict[SingleMemcachedInstanceEndpoint, asyncio.Task[None]] = {}
        self.__endpoint_healthcheck_config: EndpointHealthcheckConfig = (
            endpoint_healthcheck_config or EndpointHealthcheckConfig()
        )
        self.__autodiscovery_task: asyncio.Task[None] | None = None

    @property
    def metrics(self) -> PoolMetrics:
        """
        Aggregate metrics obtained from the sub-pools for each
        endpoint that this cluster pool is configured against.
        """
        return PoolMetrics.merge(
            [
                self._cluster_pools[endpoint].metrics
                for endpoint in self.__all_endpoints
                if endpoint in self._cluster_pools
            ]
        )

    @property
    def endpoints(self) -> set[SingleMemcachedInstanceEndpoint]:
        return self.__all_endpoints - self.__unhealthy_endpoints

    async def __autodiscovery_query(self) -> None:
        if self.__autodiscovery_endpoint:
            with closing(
                TCPConnection(
                    (self.__autodiscovery_endpoint.host, self.__autodiscovery_endpoint.port),
                    connect_timeout=self._connection_parameters.get(
                        "connect_timeout", CONNECT_TIMEOUT
                    ),
                    read_timeout=self._connection_parameters.get("read_timeout", READ_TIMEOUT),
                )
            ) as connection:
                await connection.connect()
                command = AWSAutoDiscoveryConfig()
                connection.create_request(command)
                autodiscovery_version, endpoints = await command.response
                if autodiscovery_version > self.__autodiscovery_current_version:
                    new_endpoints = endpoints - self.__all_endpoints
                    discarded_endpoints = self.__all_endpoints - endpoints
                    for endpoint in new_endpoints:
                        self.add_endpoint(endpoint)
                    for endpoint in discarded_endpoints:
                        self.remove_endpoint(endpoint)
                    self.__autodiscovery_current_version = autodiscovery_version

    async def __refresh_autodiscovered_endpoints(self) -> None:
        if not self.__autodiscovery_endpoint:
            return
        while True:
            try:
                await self.__autodiscovery_query()
                await asyncio.sleep(self.__autodiscovery_endpoint.refresh_interval)
            except asyncio.CancelledError:
                break

    async def initialize(self) -> None:
        if self.__initialized:
            return
        async with self.__pool_lock:
            if self.__initialized:
                return
            if self.__autodiscovery_endpoint:
                await self.__autodiscovery_query()
                self.__autodiscovery_task = asyncio.create_task(
                    self.__refresh_autodiscovered_endpoints()
                )
            else:
                for endpoint in self.endpoints:
                    self.add_endpoint(endpoint)
            await asyncio.gather(
                *[self._cluster_pools[endpoint].initialize() for endpoint in self.endpoints]
            )
            self.__initialized = True

    async def execute_command(self, command: Command[R]) -> None:
        """
        Dispatches a command to the appropriate memcached endpoint(s).
        To receive the response the future pointed to by ``command.response`` should be awaited
        as it will be updated when the response(s) (or exception) are available on the transport
        and merged (if it is a command that spans multiple endpoints).
        """
        try:
            await self.initialize()
            if command.keys and len(command.keys) == 1:
                await self._cluster_pools[self._router.get_node(command.keys[0])].execute_command(
                    command
                )
            else:
                mapping = defaultdict(list)
                if command.keys:
                    for key in command.keys:
                        mapping[self._router.get_node(key)].append(key)
                    endpoint_commands = {
                        endpoint: command.clone(keys) for endpoint, keys in mapping.items()
                    }
                else:
                    endpoint_commands = {
                        endpoint: command.clone(command.keys) for endpoint in self.endpoints
                    }
                await asyncio.gather(
                    *[
                        self._cluster_pools[endpoint].execute_command(endpoint_command)
                        for endpoint, endpoint_command in endpoint_commands.items()
                    ]
                )
                if not command.noreply:
                    command.response.set_result(
                        command.merge(
                            await asyncio.gather(
                                *[command.response for command in endpoint_commands.values()]
                            )
                        )
                    )
        except MemcachioConnectionError as e:
            if self.__endpoint_healthcheck_config.remove_unhealthy_endpoints:
                if (
                    not (current_task := self.__healthcheck_tasks.get(e.endpoint, None))
                    or current_task.done()
                ):
                    self.__healthcheck_tasks[e.endpoint] = asyncio.create_task(
                        self.__check_endpoint_health(e.endpoint)
                    )
            raise

    async def __check_endpoint_health(self, endpoint: SingleMemcachedInstanceEndpoint) -> None:
        attempt = 0
        while True:
            try:
                try:
                    if pool := self._cluster_pools.get(endpoint, None):
                        if (
                            pool.metrics.connection_errors
                            < self.__endpoint_healthcheck_config.maximum_error_count_for_removal
                        ):
                            return
                        await pool.initialize()
                        if pool.metrics.current_connections > 0:
                            if self.__endpoint_healthcheck_config.monitor_unhealthy_endpoints:
                                logger.info(
                                    f"Memcached server at {endpoint} has recovered after {2**attempt} seconds"
                                )
                                self.update_endpoint_status(endpoint, EndpointStatus.UP)
                            break
                        else:
                            pool.close()
                except MemcachioConnectionError:
                    self.update_endpoint_status(endpoint, EndpointStatus.DOWN)
                    if (
                        not self.__endpoint_healthcheck_config.monitor_unhealthy_endpoints
                        or attempt == self.__endpoint_healthcheck_config.maximum_recovery_attempts
                    ):
                        logger.error(f"Memcached server at {endpoint} unreachable and marked down")
                        break
                except Exception:
                    logger.exception("Unknown error while checking endpoint health")
                    break

                if (
                    endpoint in self.__unhealthy_endpoints
                    and self.__endpoint_healthcheck_config.monitor_unhealthy_endpoints
                    and attempt < self.__endpoint_healthcheck_config.maximum_recovery_attempts
                ):
                    match self.__endpoint_healthcheck_config.retry_backoff_policy:
                        case "linear":
                            delay = attempt
                        case "exponential":
                            delay = 2**attempt
                    logger.debug(
                        f"Memcached server at {endpoint} still down, attempting recovery attempt in {delay} seconds"
                    )
                    attempt += 1
                    await asyncio.sleep(delay)
            except asyncio.CancelledError:
                break

    def close(self) -> None:
        for pool in self._cluster_pools.values():
            pool.close()
        for task in self.__healthcheck_tasks.values():
            task.cancel()
        if self.__autodiscovery_task:
            self.__autodiscovery_task.cancel()
            self.__autodiscovery_task = None
        self.__healthcheck_tasks.clear()
        self.__unhealthy_endpoints.clear()
        self.__initialized = False

    def add_endpoint(self, endpoint: SingleMemcachedInstanceEndpoint) -> None:
        """
        Add a new endpoint to this pool
        """
        normalized_endpoint = normalize_single_server_endpoint(endpoint)
        self.__all_endpoints.add(normalized_endpoint)
        self._router.add_node(normalized_endpoint)
        if normalized_endpoint not in self._cluster_pools:
            self._cluster_pools[normalized_endpoint] = SingleServerPool(
                normalized_endpoint,
                min_connections=self._min_connections,
                max_connections=self._max_connections,
                blocking_timeout=self._blocking_timeout,
                idle_connection_timeout=self._idle_connection_timeout,
                **self._connection_parameters,
            )

    def remove_endpoint(self, endpoint: SingleMemcachedInstanceEndpoint) -> None:
        """
        Remove an endpoint from this pool. This will immediately also close
        all connections to that endpoint.
        """
        normalized_endpoint = normalize_single_server_endpoint(endpoint)
        self.__all_endpoints.discard(normalized_endpoint)
        self._router.remove_node(normalized_endpoint)
        if pool := self._cluster_pools.pop(normalized_endpoint, None):
            pool.close()

    def update_endpoint_status(
        self, endpoint: SingleMemcachedInstanceEndpoint, status: EndpointStatus
    ) -> None:
        """
        Change the status of an endpoint in this pool.
        Marking an endpoint as :enum:`EndpointStatus.DOWN` will immediately stop routing
        requests to it, while marking it as :enum:`EndpointStatus.UP` will immediately
        start routing requests to it.
        """
        normalized_endpoint = normalize_single_server_endpoint(endpoint)
        match status:
            case EndpointStatus.UP:
                self.__unhealthy_endpoints.discard(normalized_endpoint)
                self._router.add_node(normalized_endpoint)
            case EndpointStatus.DOWN:
                self.__unhealthy_endpoints.add(normalized_endpoint)
                self._router.remove_node(normalized_endpoint)
        self._cluster_pools[normalized_endpoint].metrics.on_status_update(status)



from __future__ import annotations

import asyncio
import dataclasses
import socket
import time
import weakref
from abc import ABC, abstractmethod
from asyncio import (
    BaseProtocol,
    BaseTransport,
    Event,
    Future,
    Lock,
    Transport,
    get_running_loop,
)
from collections import deque
from collections.abc import Callable
from contextlib import suppress
from io import BytesIO
from pathlib import Path
from ssl import SSLContext
from typing import Any, Generic, TypedDict, TypeVar, cast

from .authentication import Authenticator, SimpleAuthenticator
from .commands import Command
from .compat import NotRequired, Unpack, asyncio_timeout
from .defaults import (
    CONNECT_TIMEOUT,
    MAX_AVERAGE_RESPONSE_TIME_FOR_CONNECTION_REUSE,
    MAX_INFLIGHT_REQUESTS_PER_CONNECTION,
    READ_TIMEOUT,
)
from .errors import MemcachedError, MemcachioConnectionError, NotEnoughData
from .types import SingleMemcachedInstanceEndpoint, TCPEndpoint, UnixSocketEndpoint
from .utils import decodedstr

R = TypeVar("R")


class ConnectionParams(TypedDict):
    #: Maximum time to wait when establishing a connection
    connect_timeout: float | None
    #: Maximum time to wait for a response
    read_timeout: float | None
    #: Whether to set the :data:`socket.TCP_NODELAY` flag on the socket
    socket_nodelay: NotRequired[bool | None]
    #: Whether to set the :data:`socket.SO_KEEPALIVE` flag on the socket
    socket_keepalive: NotRequired[bool | None]
    #: Additional options to set if ``socket_keepalive`` is ``True``
    socket_keepalive_options: NotRequired[dict[int, int | bytes] | None]
    #: Maximum number of concurrent requests to allow on the connection
    max_inflight_requests_per_connection: NotRequired[int]
    #:  Threshold for allowing the connection to be reused when there are requests pending.
    max_average_response_time_for_connection_reuse: NotRequired[float]
    #: SSL context to use if connecting to a memcached instance listening on a secure port
    ssl_context: NotRequired[SSLContext | None]
    #: Authentication strategy to use after establishing a connection
    authenticator: NotRequired[Authenticator | None]
    #: Username for SASL authentication
    username: NotRequired[str | None]
    #: Password for SASL authentication
    password: NotRequired[str | None]
    #: List of callables that will be called with the connection as the first argument
    #  when the connection is successfully established.
    on_connect_callbacks: NotRequired[list[Callable[[BaseConnection], None]]]
    #: List of callables that will be called with the connection as the first argument
    #  when the connection is disconnected from the server
    on_disconnect_callbacks: NotRequired[list[Callable[[BaseConnection], None]]]


@dataclasses.dataclass
class Request(Generic[R]):
    connection: weakref.ReferenceType[BaseConnection]
    command: Command[R]
    decode: bool = False
    encoding: str | None = None
    raise_exceptions: bool = True
    created_at: float = dataclasses.field(default_factory=lambda: time.time())
    timeout_handler: asyncio.Handle | None = None

    def __post_init__(self) -> None:
        if connection := self.connection():
            connection.metrics.requests_pending += 1
        self.command.response.add_done_callback(self.__update_metrics)
        self.command.response.add_done_callback(self.__cancel_timer)

    def __cancel_timer(self, future: Future[R]) -> None:
        if (
            self.timeout_handler
            and future.done()
            and not (future.cancelled() or future.exception())
        ):
            self.timeout_handler.cancel()

    def __update_metrics(self, future: Future[R]) -> None:
        if connection := self.connection():
            metrics = connection.metrics
            metrics.last_request_processed = time.time()
            metrics.requests_pending -= 1
            if future.done() and not future.cancelled():
                if not future.exception():
                    metrics.requests_processed += 1
                    metrics.average_response_time = (
                        (time.time() - self.created_at)
                        + metrics.average_response_time * (metrics.requests_processed - 1)
                    ) / metrics.requests_processed
                else:
                    metrics.requests_failed += 1


@dataclasses.dataclass
class ConnectionMetrics:
    """
    Tracks metrics for a connection.
    """

    #: Timestamp when the connection was established.
    created_at: float | None = None
    #: Total number of successfully processed requests.
    requests_processed: int = 0
    #: Total number of requests that failed.
    requests_failed: int = 0
    #: Timestamp when data was last written.
    last_written: float = 0.0
    #: Timestamp when data was last read.
    last_read: float = 0.0
    #: Timestamp when the last request completed processing.
    last_request_processed: float = 0.0
    #: Average time taken to process requests.
    average_response_time: float = 0.0
    #: Number of requests currently pending.
    requests_pending: int = 0


class BaseConnection(BaseProtocol, ABC):
    """Wraps an asyncio connection using a custom protocol.
    Provides methods for sending commands and reading lines.
    """

    endpoint: SingleMemcachedInstanceEndpoint
    #: Metrics related to this connection
    metrics: ConnectionMetrics

    def __init__(
        self,
        *,
        connect_timeout: float | None = CONNECT_TIMEOUT,
        read_timeout: float | None = READ_TIMEOUT,
        socket_keepalive: bool | None = True,
        socket_keepalive_options: dict[int, int | bytes] | None = None,
        socket_nodelay: bool | None = False,
        max_inflight_requests_per_connection: int = MAX_INFLIGHT_REQUESTS_PER_CONNECTION,
        max_average_response_time_for_connection_reuse: float = MAX_AVERAGE_RESPONSE_TIME_FOR_CONNECTION_REUSE,
        ssl_context: SSLContext | None = None,
        authenticator: Authenticator | None = None,
        username: str | None = None,
        password: str | None = None,
        on_connect_callbacks: list[Callable[[BaseConnection], None]] | None = None,
        on_disconnect_callbacks: list[Callable[[BaseConnection], None]] | None = None,
    ) -> None:
        """
        :param connect_timeout: Timeout for establishing a connection.
        :param read_timeout: Timeout for reading data from the connection.
        :param socket_keepalive: Whether to enable SO_KEEPALIVE on the socket.
        :param socket_keepalive_options: Additional keepalive options.
        :param socket_nodelay: Whether to enable TCP_NODELAY on the socket.
        :param max_inflight_requests_per_connection: Maximum concurrent requests allowed.
        :param max_average_response_time_for_connection_reuse: Threshold for allowing the connection to be
         reused when there are requests pending.
        :param ssl_context: SSL context for secure connections.
        :param authenticator: Authentication strategy to use after establishing a connection
        :param username: Username for SASL authentication.
        :param password: Password for SASL authentication.
        :param on_connect_callbacks: Callbacks to invoke upon successful connection.
        :param on_disconnect_callbacks: Callbacks to invoke upon disconnection.
        """
        self._connect_timeout: float | None = connect_timeout
        self._read_timeout: float | None = read_timeout
        self._socket_nodelay: bool | None = socket_nodelay
        self._socket_keepalive: bool | None = socket_keepalive
        self._socket_keepalive_options: dict[int, int | bytes] = socket_keepalive_options or {}
        self._max_inflight_requests_per_connection = max_inflight_requests_per_connection
        self._max_average_response_time_for_connection_reuse = (
            max_average_response_time_for_connection_reuse
        )
        self._ssl_context: SSLContext | None = ssl_context
        self._last_error: Exception | None = None
        self._transport: Transport | None = None
        self._buffer = BytesIO()
        self._request_queue: deque[Request[Any]] = deque()
        self._write_ready: Event = Event()
        self._transport_lock: Lock = Lock()
        self._request_lock: Lock = Lock()
        self._connect_callbacks = [weakref.proxy(cb) for cb in on_connect_callbacks or []]
        self._disconnect_callbacks = [weakref.proxy(cb) for cb in on_disconnect_callbacks or []]
        self.metrics: ConnectionMetrics = ConnectionMetrics()
        self._authenticator = authenticator or (
            SimpleAuthenticator(username, password) if (username and password) else None
        )

    @abstractmethod
    async def connect(self) -> None:
        """
        Establish a connection to the target memcached server
        """
        ...

    @property
    def connected(self) -> bool:
        """
        Whether the connection is currently connected
        """
        return self._transport is not None and self._write_ready.is_set()

    @property
    def reusable(self) -> bool:
        """
        Whether this connection is healthy enough to handle
        any more concurrent requests
        """
        return (
            len(self._request_queue) < self._max_inflight_requests_per_connection
            and self.metrics.average_response_time
            < self._max_average_response_time_for_connection_reuse
        )

    def send(self, data: bytes) -> None:
        assert self._transport
        self._transport.write(data)
        self.metrics.last_written = time.time()

    def close(self) -> None:
        """
        Disconnect from the memcached server and clear internal
        state.
        """
        self.__on_disconnect()

    def create_request(self, command: Command[R]) -> None:
        """
        Send a request to the memcached server and queue the response
        handling if this is not a ``noreply`` request.
        """
        self.send(bytes(command.build_request()))
        command.request_sent.set_result(True)
        if not command.noreply:
            request = Request(
                weakref.ref(self),
                command,
            )
            self._request_queue.append(request)
            if self._read_timeout is not None:
                request.timeout_handler = asyncio.get_running_loop().call_later(
                    self._read_timeout,
                    lambda command: command.response.set_exception(
                        TimeoutError(
                            f"command {decodedstr(command.name)} timed out after {self._read_timeout} seconds"
                        )
                    )
                    if not command.response.done()
                    else None,
                    command,
                )

    def connection_made(self, transport: BaseTransport) -> None:
        """
        :meta private:
        """
        if self.metrics.created_at:
            self.metrics = ConnectionMetrics()

        self.metrics.created_at = time.time()
        self._transport = cast(Transport, transport)
        if (sock := self._transport.get_extra_info("socket")) is not None:
            try:
                if self._socket_nodelay is not None and sock.family in (
                    socket.AF_INET,
                    socket.AF_INET6,
                ):
                    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, self._socket_nodelay)

                if self._socket_keepalive is not None:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, self._socket_keepalive)
                    for k, v in self._socket_keepalive_options.items():
                        sock.setsockopt(socket.IPPROTO_TCP, k, v)
            except (OSError, TypeError) as err:
                transport.close()
                self._last_error = err
                return
        with suppress(RuntimeError):
            [cb(self) for cb in self._connect_callbacks]
        self._write_ready.set()

    def data_received(self, data: bytes) -> None:
        """:meta private:"""
        self.metrics.last_read = time.time()
        self._buffer = BytesIO(self._buffer.read() + data)
        while self._request_queue:
            request = self._request_queue.popleft()
            try:
                response = request.command.parse(self._buffer, self.endpoint)
                if not (request.command.response.cancelled() or request.command.response.done()):
                    request.command.response.set_result(response)
            except NotEnoughData as e:
                self._buffer.seek(self._buffer.tell() - e.data_read)
                self._request_queue.appendleft(request)
                break
            except MemcachedError as e:
                if not (request.command.response.cancelled() or request.command.response.done()):
                    request.command.response.set_exception(e)
            except Exception as e:
                self._request_queue.appendleft(request)
                self._last_error = e
                break

    def pause_writing(self) -> None:
        """:meta private:"""
        self._write_ready.clear()

    def resume_writing(self) -> None:
        """:meta private:"""
        self._write_ready.set()

    def connection_lost(self, exc: Exception | None) -> None:
        """
        :meta private:
        """
        if exc:
            self._last_error = exc
        self.__on_disconnect(True, "Connection lost")

    def eof_received(self) -> None:
        """:meta private:"""
        self.__on_disconnect(True, "EOF received")

    async def _authenticate(self) -> None:
        if self._authenticator:
            await self._authenticator.authenticate(self)

    def __on_disconnect(self, from_server: bool = False, reason: str | None = None) -> None:
        self._write_ready.clear()
        if self._transport:
            try:
                self._transport.close()
            except RuntimeError:
                pass
            self._transport = None

        while True:
            try:
                request = self._request_queue.popleft()
                if not request.command.response.done():
                    exc = MemcachioConnectionError(reason or "", self.endpoint)
                    if self._last_error:
                        exc.__cause__ = self._last_error
                    request.command.response.set_exception(exc)
            except IndexError:
                break

        self._buffer = BytesIO()

        if from_server:
            with suppress(RuntimeError):
                [cb(self) for cb in self._disconnect_callbacks]


class TCPConnection(BaseConnection):
    def __init__(
        self,
        host_port: TCPEndpoint | tuple[str, int],
        **kwargs: Unpack[ConnectionParams],
    ) -> None:
        self._host, self._port = host_port
        self.endpoint = host_port
        super().__init__(**kwargs)

    async def connect(self) -> None:
        """
        Establish a connection to the target memcached server listening on
        a tcp port
        """
        if not self._transport:
            async with self._transport_lock:
                if not self._transport:
                    try:
                        async with asyncio_timeout(self._connect_timeout):
                            transport, _ = await get_running_loop().create_connection(
                                lambda: self,
                                host=self._host,
                                port=self._port,
                                ssl=self._ssl_context,
                            )
                            await self._write_ready.wait()
                            await self._authenticate()
                    except (OSError, asyncio.TimeoutError) as e:
                        msg = f"Unable to establish a connection within {self._connect_timeout} seconds"
                        raise MemcachioConnectionError(msg, self.endpoint) from (
                            self._last_error or e
                        )


class UnixSocketConnection(BaseConnection):
    def __init__(
        self,
        path: UnixSocketEndpoint,
        **kwargs: Unpack[ConnectionParams],
    ) -> None:
        self.endpoint = self._path = str(Path(path).expanduser().absolute())
        super().__init__(**kwargs)

    async def connect(self) -> None:
        """
        Establish a connection to the target memcached server listening on
        a unix domain socket
        """
        if not self._transport:
            async with self._transport_lock:
                if not self._transport:
                    try:
                        async with asyncio_timeout(self._connect_timeout):
                            transport, _ = await get_running_loop().create_unix_connection(
                                lambda: self, path=self._path
                            )
                    except (OSError, asyncio.TimeoutError) as e:
                        msg = f"Unable to establish a connection within {self._connect_timeout} seconds"
                        raise MemcachioConnectionError(msg, self.endpoint) from e
                    await self._write_ready.wait()
                    await self._authenticate()



"""memcachio

async memcached client
"""

from __future__ import annotations

from . import _version, defaults, errors
from .authentication import Authenticator, SimpleAuthenticator
from .client import Client
from .connection import (
    BaseConnection,
    ConnectionMetrics,
    ConnectionParams,
    TCPConnection,
    UnixSocketConnection,
)
from .pool import (
    ClusterPool,
    EndpointHealthcheckConfig,
    EndpointStatus,
    Pool,
    PoolMetrics,
    SingleServerPool,
)
from .types import (
    AWSAutoDiscoveryEndpoint,
    MemcachedEndpoint,
    MemcachedItem,
    TCPEndpoint,
    UnixSocketEndpoint,
)

__all__ = [
    "Authenticator",
    "AWSAutoDiscoveryEndpoint",
    "BaseConnection",
    "Client",
    "ClusterPool",
    "ConnectionParams",
    "ConnectionMetrics",
    "MemcachedItem",
    "MemcachedEndpoint",
    "EndpointStatus",
    "EndpointHealthcheckConfig",
    "Pool",
    "PoolMetrics",
    "SimpleAuthenticator",
    "SingleServerPool",
    "TCPConnection",
    "TCPEndpoint",
    "UnixSocketConnection",
    "UnixSocketEndpoint",
    "defaults",
    "errors",
]
__version__ = _version.get_versions()["version"]



from __future__ import annotations

import abc
import asyncio
import copy
import dataclasses
import time
import weakref
from asyncio import Future
from collections import ChainMap
from collections.abc import Sequence
from io import BytesIO
from typing import AnyStr, ClassVar, Generic, TypeVar, cast

from .compat import Self
from .constants import LINE_END, Commands, Responses
from .errors import (
    AutoDiscoveryError,
    ClientError,
    MemcachedError,
    NotEnoughData,
    ServerError,
)
from .types import KeyT, MemcachedItem, SingleMemcachedInstanceEndpoint, TCPEndpoint, ValueT
from .utils import bytestr, decodedstr

R = TypeVar("R")


@dataclasses.dataclass
class Request(Generic[R]):
    command: weakref.ProxyType[Command[R]]
    header: bytes
    body: list[bytes] = dataclasses.field(default_factory=lambda: [])

    def __bytes__(self) -> bytes:
        request_body: bytes = self.command.name
        if self.header:
            request_body += b" %b" % self.header
        if self.command.noreply:
            request_body += b" noreply"
        request_body += LINE_END
        request_body += LINE_END.join(self.body)
        return request_body


class Command(abc.ABC, Generic[R]):
    """
    The abstract generic command class used throughout the
    client -> connection pool -> connection lifecycle to
    manage dispatching a request, parsing the response (if not
    a ``noreply`` request) and resolving the future to receive
    the eventual parsed response.

    This class is not meant for direct use, however it's structure
    might be interesting if implementing a custom connection pool or
    using the connection classes directly.

    As an example the following snippet demonstrates creating a
    ``set`` command that takes no flags or expiry and is dispatched
    directly to a connection without using :class:`~memcachio.Client`
    or a :class:`~memcachio.pool.Pool`::

        import asyncio
        import weakref
        from io import BytesIO

        import memcachio
        from memcachio.commands import Command, Request
        from memcachio.constants import Commands


        class MyCustomSetCommand(Command[bool]):
            name = Commands.SET

            def __init__(self, key: str, value: bytes, noreply: bool = False):
                self.value = value
                super().__init__(key, noreply=noreply)

            def build_request(self) -> memcachio.commands.Request[bool]:
                return Request(
                    weakref.proxy(self),
                    b"%b 0 0 %d" % (self.keys[0].encode(), len(self.value)),
                    [self.value + b"\\r\\n"],
                )

            def parse(self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint) -> bool:
                response = data.readline()
                return response.rstrip() == b"STORED"


        async def example():
            command = MyCustomSetCommand("fubar", b"1234")
            connection = memcachio.TCPConnection(("localhost", 11211))
            await connection.connect()
            connection.create_request(command)
            assert await command.response


        asyncio.run(example())

    """

    __slots__ = (
        "__weakref__",
        "_keys",
        "noreply",
        "request_sent",
        "response",
        "response_time",
        "created_at",
    )

    #: The name of the command
    name: ClassVar[Commands]
    #: A future that should be set when the request has been written
    #:  to the socket
    request_sent: Future[bool]
    #: A future that should be set when the request has received a response
    #:  and parsed
    response: Future[R]
    #:
    response_time: float

    def __init__(self, *keys: KeyT, noreply: bool = False):
        """
        :param keys:  The keys this command operates on
        :param noreply: Whether the server should send a response
         to the command.
        """
        self.noreply = noreply
        self.created_at = time.time()
        self.response_time = 0
        self._keys: list[str] = [decodedstr(key) for key in keys or []]
        self.request_sent = asyncio.get_running_loop().create_future()
        self.response = asyncio.get_running_loop().create_future()
        (self.request_sent if self.noreply else self.response).add_done_callback(
            lambda _: self._update_response_time()
        )

    def merge(self, responses: list[R]) -> R:
        return responses[0]

    @property
    def keys(self) -> list[str]:
        return self._keys

    def clone(self, keys: Sequence[KeyT]) -> Self:
        subset = copy.copy(self)
        subset._keys = list(decodedstr(key) for key in keys)
        subset.request_sent = asyncio.get_running_loop().create_future()
        subset.response = asyncio.get_running_loop().create_future()
        (subset.request_sent if subset.noreply else subset.response).add_done_callback(
            lambda _: subset._update_response_time()
        )
        return subset

    @abc.abstractmethod
    def build_request(self) -> Request[R]:
        """
        Build the header and (optional) payload for the request to
        send to the memcached server
        """
        ...

    @abc.abstractmethod
    def parse(self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint) -> R:
        """
        Parse the response from the buffer assuming the
        current position of the buffer contains the response
        for this command.
        """
        ...

    def _update_response_time(self) -> None:
        self.response_time = time.time() - self.created_at

    def _check_header(self, header: bytes) -> None:
        if not header.endswith(LINE_END):
            raise NotEnoughData(len(header))
        response = header.rstrip()
        if response.startswith(Responses.CLIENT_ERROR):
            raise ClientError(decodedstr(response.split(Responses.CLIENT_ERROR)[1]).strip())
        elif response.startswith(Responses.SERVER_ERROR):
            raise ServerError(decodedstr(response.split(Responses.SERVER_ERROR)[1]).strip())
        elif response.startswith(Responses.ERROR):
            raise MemcachedError(decodedstr(response).strip())
        return None


class BasicResponseCommand(Command[bool]):
    success: ClassVar[Responses]

    def parse(self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint) -> bool:
        response = data.readline()
        self._check_header(response)
        if not response.rstrip() == self.success.value:
            return False
        return True


class GetCommand(Command[dict[AnyStr, MemcachedItem[AnyStr]]]):
    __slots__ = ("decode_responses", "encoding", "items")
    name = Commands.GET

    def __init__(self, *keys: KeyT, decode: bool = False, encoding: str = "utf-8") -> None:
        self.items: list[MemcachedItem[AnyStr]] = []
        self.decode_responses = decode
        self.encoding = encoding
        super().__init__(*keys, noreply=False)

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), f"{' '.join(self.keys)}".encode())

    def parse(
        self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint
    ) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        while True:
            header = data.readline()
            self._check_header(header)
            if header.lstrip() == (Responses.END + LINE_END):
                break
            parts = header.split()
            if len(parts) < 4 or parts[0] != Responses.VALUE:
                msg = f"Unexpected response header: {decodedstr(header)}"
                raise ValueError(msg)
            key = parts[1]
            flags = int(parts[2])
            size = int(parts[3])
            cas = int(parts[4]) if len(parts) > 4 else None
            value = data.read(size)
            if len(value) != size:
                raise NotEnoughData(len(value) + len(header))
            item = MemcachedItem[AnyStr](
                cast(AnyStr, decodedstr(key, self.encoding) if self.decode_responses else key),
                flags,
                size,
                cas,
                cast(AnyStr, decodedstr(value, self.encoding) if self.decode_responses else value),
            )
            data.read(2)
            self.items.append(item)
        return {i.key: i for i in self.items}

    def merge(
        self, results: list[dict[AnyStr, MemcachedItem[AnyStr]]]
    ) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        merged = {}
        for res in results:
            for key, item in res.items():
                merged[key] = item
        return merged


class GetsCommand(GetCommand[AnyStr]):
    name = Commands.GETS


class GatCommand(GetCommand[AnyStr]):
    __slots__ = ("expiry",)
    name = Commands.GAT

    def __init__(
        self,
        *keys: KeyT,
        expiry: int = 0,
        decode: bool = False,
        encoding: str = "utf-8",
    ) -> None:
        self.expiry = expiry
        super().__init__(*keys, decode=decode, encoding=encoding)

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), f"{self.expiry} {' '.join(self.keys)}".encode())


class GatsCommand(GatCommand[AnyStr]):
    name = Commands.GATS


class GenericStoreCommand(BasicResponseCommand):
    __slots__ = ("cas", "encoding", "expiry", "flags", "value")

    def __init__(
        self,
        key: KeyT,
        value: ValueT,
        *,
        flags: int | None = None,
        expiry: int = 0,
        noreply: bool = False,
        cas: int | None = None,
        encoding: str = "utf-8",
    ) -> None:
        self.encoding = encoding
        self.flags = flags
        self.expiry = expiry
        self.value = bytestr(value, self.encoding)
        self.cas = cas
        super().__init__(key, noreply=noreply)

    def build_request(self) -> Request[R]:
        header = f"{decodedstr(self.keys[0])} {self.flags or 0} {self.expiry}"
        header += f" {len(self.value)}"
        if self.cas is not None:
            header += f" {self.cas}"
        return Request(weakref.proxy(self), header.encode(), [self.value + LINE_END])


class SetCommand(GenericStoreCommand):
    name = Commands.SET
    success = Responses.STORED


class CheckAndSetCommand(GenericStoreCommand):
    name = Commands.CAS
    success = Responses.STORED


class AddCommand(GenericStoreCommand):
    name = Commands.ADD
    success = Responses.STORED


class AppendCommand(GenericStoreCommand):
    name = Commands.APPEND
    success = Responses.STORED


class PrependCommand(GenericStoreCommand):
    name = Commands.PREPEND
    success = Responses.STORED


class ReplaceCommand(GenericStoreCommand):
    name = Commands.REPLACE
    success = Responses.STORED


class ArithmenticCommand(Command[int | None]):
    __slots__ = ("amount",)

    def __init__(self, key: KeyT, amount: int, noreply: bool) -> None:
        self.amount = amount
        super().__init__(key, noreply=noreply)

    def build_request(self) -> Request[R]:
        request = f"{decodedstr(self.keys[0])} {self.amount}"
        return Request(weakref.proxy(self), request.encode())

    def parse(self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint) -> int | None:
        response = data.readline()
        self._check_header(response)
        response = response.rstrip()
        if response == Responses.NOT_FOUND:
            return None
        return int(response)


class IncrCommand(ArithmenticCommand):
    name = Commands.INCR


class DecrCommand(ArithmenticCommand):
    name = Commands.DECR


class DeleteCommand(BasicResponseCommand):
    name = Commands.DELETE
    success = Responses.DELETED

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), bytestr(self.keys[0]))


class TouchCommand(BasicResponseCommand):
    __slots__ = ("expiry",)
    name = Commands.TOUCH
    success = Responses.TOUCHED

    def __init__(self, key: KeyT, *, expiry: int, noreply: bool = False) -> None:
        self.expiry = expiry
        super().__init__(key, noreply=noreply)

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), f"{self.keys[0]} {self.expiry}".encode())


class FlushAllCommand(BasicResponseCommand):
    __slots__ = ("expiry",)
    name = Commands.FLUSH_ALL
    success = Responses.OK

    def __init__(self, expiry: int) -> None:
        self.expiry = expiry
        super().__init__()

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), bytestr(self.expiry))

    def merge(self, results: list[bool]) -> bool:
        return all(results)


class VersionCommand(Command[dict[SingleMemcachedInstanceEndpoint, str]]):
    name = Commands.VERSION

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), b"")

    def parse(
        self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint
    ) -> dict[SingleMemcachedInstanceEndpoint, str]:
        response = data.readline()
        self._check_header(response)
        return {endpoint: response.partition(Responses.VERSION)[-1].strip().decode("utf-8")}

    def merge(
        self, responses: list[dict[SingleMemcachedInstanceEndpoint, str]]
    ) -> dict[SingleMemcachedInstanceEndpoint, str]:
        return dict(ChainMap(*responses))


class StatsCommand(Command[dict[SingleMemcachedInstanceEndpoint, dict[AnyStr, AnyStr]]]):
    name = Commands.STATS

    def __init__(
        self, arg: str | None = None, *, decode_responses: bool = False, encoding: str = "utf-8"
    ):
        self.arg = arg
        self.decode_responses = decode_responses
        self.encoding = encoding
        super().__init__(noreply=False)

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), b"" if not self.arg else bytestr(self.arg))

    def parse(
        self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint
    ) -> dict[SingleMemcachedInstanceEndpoint, dict[AnyStr, AnyStr]]:
        stats = {}
        while True:
            section = data.readline()
            self._check_header(section)
            if section.startswith(Responses.END):
                break
            elif section.startswith(Responses.STAT):
                part = section.lstrip(Responses.STAT).strip()
                item, value = (decodedstr(part) if self.decode_responses else part).split()
                stats[cast(AnyStr, item)] = cast(AnyStr, value)
        return {endpoint: stats}

    def merge(
        self, responses: list[dict[SingleMemcachedInstanceEndpoint, dict[AnyStr, AnyStr]]]
    ) -> dict[SingleMemcachedInstanceEndpoint, dict[AnyStr, AnyStr]]:
        return dict(ChainMap(*responses))


class AWSAutoDiscoveryConfig(Command[tuple[int, set[SingleMemcachedInstanceEndpoint]]]):
    name = Commands.CONFIG

    def build_request(self) -> Request[R]:
        return Request(weakref.proxy(self), b"get cluster")

    def parse(
        self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint
    ) -> tuple[int, set[SingleMemcachedInstanceEndpoint]]:
        header = data.readline()
        self._check_header(header)
        version = int(data.readline())
        parsed_endpoints = [host.split(b"|") for host in data.readline().strip().split(b" ")]
        data.readline()
        if (data.readline().strip()) != Responses.END:
            raise AutoDiscoveryError("Malformed response")
        return version, {TCPEndpoint(host[1].decode(), int(host[2])) for host in parsed_endpoints}



from __future__ import annotations

from collections.abc import Callable
from ssl import SSLContext
from typing import (
    AnyStr,
    Generic,
    Literal,
    ParamSpec,
    TypeVar,
    overload,
)

from .authentication import Authenticator
from .commands import (
    AddCommand,
    AppendCommand,
    CheckAndSetCommand,
    Command,
    DecrCommand,
    DeleteCommand,
    FlushAllCommand,
    GatCommand,
    GatsCommand,
    GetCommand,
    GetsCommand,
    IncrCommand,
    PrependCommand,
    ReplaceCommand,
    SetCommand,
    StatsCommand,
    TouchCommand,
    VersionCommand,
)
from .compat import Unpack
from .connection import ConnectionParams
from .defaults import (
    BLOCKING_TIMEOUT,
    CONNECT_TIMEOUT,
    ENCODING,
    IDLE_CONNECTION_TIMEOUT,
    MAX_AVERAGE_RESPONSE_TIME_FOR_CONNECTION_REUSE,
    MAX_CONNECTIONS,
    MAX_INFLIGHT_REQUESTS_PER_CONNECTION,
    MIN_CONNECTIONS,
    READ_TIMEOUT,
)
from .pool import ClusterPool, EndpointHealthcheckConfig, Pool, SingleServerPool
from .types import (
    KeyT,
    MemcachedEndpoint,
    MemcachedItem,
    SingleMemcachedInstanceEndpoint,
    ValueT,
    is_single_server,
)

R = TypeVar("R")
P = ParamSpec("P")


class Client(Generic[AnyStr]):
    connection_pool: Pool

    @overload
    def __init__(
        self: Client[str],
        memcached_location: MemcachedEndpoint | None = ...,
        decode_responses: Literal[True] = True,
        encoding: str = ...,
        min_connections: int = ...,
        max_connections: int = ...,
        blocking_timeout: float = ...,
        idle_connection_timeout: float = ...,
        hashing_function: Callable[[str], int] | None = ...,
        endpoint_healthcheck_config: EndpointHealthcheckConfig | None = ...,
        connection_pool: Pool | None = ...,
        connect_timeout: float | None = ...,
        read_timeout: float | None = ...,
        socket_nodelay: bool | None = ...,
        socket_keepalive: bool | None = ...,
        socket_keepalive_options: dict[int, int | bytes] | None = ...,
        max_inflight_requests_per_connection: int = ...,
        max_average_response_time_for_connection_reuse: float = ...,
        ssl_context: SSLContext | None = ...,
        username: str | None = ...,
        password: str | None = ...,
        authenticator: Authenticator | None = ...,
    ) -> None: ...

    @overload
    def __init__(
        self: Client[bytes],
        memcached_location: MemcachedEndpoint | None = ...,
        decode_responses: Literal[False] = False,
        encoding: str = ...,
        min_connections: int = ...,
        max_connections: int = ...,
        blocking_timeout: float = ...,
        idle_connection_timeout: float = ...,
        hashing_function: Callable[[str], int] | None = ...,
        endpoint_healthcheck_config: EndpointHealthcheckConfig | None = ...,
        connection_pool: Pool | None = ...,
        connect_timeout: float | None = ...,
        read_timeout: float | None = ...,
        socket_nodelay: bool | None = ...,
        socket_keepalive: bool | None = ...,
        socket_keepalive_options: dict[int, int | bytes] | None = ...,
        max_inflight_requests_per_connection: int = ...,
        max_average_response_time_for_connection_reuse: float = ...,
        ssl_context: SSLContext | None = ...,
        username: str | None = ...,
        password: str | None = ...,
        authenticator: Authenticator | None = ...,
    ) -> None: ...

    def __init__(
        self,
        memcached_location: MemcachedEndpoint | None = None,
        decode_responses: Literal[True, False] = False,
        encoding: str = ENCODING,
        min_connections: int = MIN_CONNECTIONS,
        max_connections: int = MAX_CONNECTIONS,
        blocking_timeout: float = BLOCKING_TIMEOUT,
        idle_connection_timeout: float = IDLE_CONNECTION_TIMEOUT,
        hashing_function: Callable[[str], int] | None = None,
        endpoint_healthcheck_config: EndpointHealthcheckConfig | None = None,
        connection_pool: Pool | None = None,
        connect_timeout: float | None = CONNECT_TIMEOUT,
        read_timeout: float | None = READ_TIMEOUT,
        socket_nodelay: bool | None = None,
        socket_keepalive: bool | None = None,
        socket_keepalive_options: dict[int, int | bytes] | None = None,
        max_inflight_requests_per_connection: int = MAX_INFLIGHT_REQUESTS_PER_CONNECTION,
        max_average_response_time_for_connection_reuse: float = MAX_AVERAGE_RESPONSE_TIME_FOR_CONNECTION_REUSE,
        ssl_context: SSLContext | None = None,
        username: str | None = None,
        password: str | None = None,
        authenticator: Authenticator | None = None,
    ) -> None:
        """
        Initialize the Memcached client.

        Either a memcached location or an existing connection pool must be provided.

        :param memcached_location: The memcached server address(es) or endpoint configuration.
        :param decode_responses: If True, decode responses using the specified encoding; otherwise, responses are returned as bytes.
        :param encoding: The character encoding used when decoding responses.
        :param max_connections: The maximum number of simultaneous connections to memcached.
        :param min_connections: The minimum number of connections to keep in the pool.
        :param blocking_timeout: The timeout (in seconds) to wait for a connection to become available.
        :param idle_connection_timeout: The maximum time to allow a connection to remain idle in the pool
         before being disconnected
        :param hashing_function: A function to use for routing keys to endpoints for multi-key
         commands. If none is provided the default :func:`hashlib.md5` implementation from the
         standard library is used.

         .. important:: This parameter is only relevant when connecting to multiple memcached servers
        :param endpoint_healthcheck_config: Configuration to control whether
         endpoints are automatically removed/recovered based on health checks.

         .. important:: This parameter is only relevant when connecting to multiple memcached servers
        :param connection_pool: An optional pre-initialized connection pool.
         If provided, ``memcached_location`` must be ``None``.

         .. caution:: All connection/connection pool related arguments will be ignored when
            a ``connection_pool`` is provided as the arguments provided when creating
            the pool itself will be in effect.
        :param connect_timeout: Timeout (in seconds) for establishing a connection.
        :param read_timeout: Timeout (in seconds) for reading from a connection.
        :param socket_nodelay: If True, disable Nagle's algorithm on the socket.
        :param socket_keepalive: If True, enable TCP keepalive on the socket.
        :param socket_keepalive_options: Additional options for configuring socket keepalive.
        :param max_inflight_requests_per_connection: Maximum number of requests allowed to be in-flight per connection.
        :param max_average_response_time_for_connection_reuse: Threshold for allowing the connection to be
         reused when there are requests pending.
        :param ssl_context: An SSL context to use for encrypted connections.
        :param username: Username for SASL authentication (if required).
        :param password: Password for SASL authentication (if required).
        :param authenticator: The authentication strategy to use when establishing new connections
        :raises ValueError: If both or neither memcached_location and connection_pool are provided.
        """
        if memcached_location and not connection_pool:
            self.connection_pool = Client.pool_from_endpoint(
                memcached_location,
                min_connections=min_connections,
                max_connections=max_connections,
                blocking_timeout=blocking_timeout,
                idle_connection_timeout=idle_connection_timeout,
                hashing_function=hashing_function,
                endpoint_healthcheck_config=endpoint_healthcheck_config,
                connect_timeout=connect_timeout,
                read_timeout=read_timeout,
                socket_nodelay=socket_nodelay,
                socket_keepalive=socket_keepalive,
                socket_keepalive_options=socket_keepalive_options,
                max_inflight_requests_per_connection=max_inflight_requests_per_connection,
                max_average_response_time_for_connection_reuse=max_average_response_time_for_connection_reuse,
                ssl_context=ssl_context,
                username=username,
                password=password,
                authenticator=authenticator,
            )
        elif connection_pool and not memcached_location:
            self.connection_pool = connection_pool
        elif connection_pool and memcached_location:
            raise ValueError(
                "One of `memcached_location` or `connection_pool` must be provided not both"
            )
        else:
            raise ValueError("One of `memcached_location` or `connection_pool` must be provided")
        self.decode_responses = decode_responses
        self.encoding = encoding

    @classmethod
    def pool_from_endpoint(
        cls,
        endpoint: MemcachedEndpoint,
        min_connections: int = MIN_CONNECTIONS,
        max_connections: int = MAX_CONNECTIONS,
        blocking_timeout: float = BLOCKING_TIMEOUT,
        idle_connection_timeout: float = IDLE_CONNECTION_TIMEOUT,
        hashing_function: Callable[[str], int] | None = None,
        endpoint_healthcheck_config: EndpointHealthcheckConfig | None = None,
        **connection_args: Unpack[ConnectionParams],
    ) -> Pool:
        """
        Returns either a :class:`~memcachio.SingleServerPool` or :class:`~memcachio.ClusterPool`
        depending on whether ``endpoint`` is a single instance or a collection of servers

        :meta private:
        """
        if is_single_server(endpoint):
            return SingleServerPool(
                endpoint,
                min_connections=min_connections,
                max_connections=max_connections,
                blocking_timeout=blocking_timeout,
                idle_connection_timeout=idle_connection_timeout,
                **connection_args,
            )
        else:
            return ClusterPool(
                endpoint,  # type: ignore[arg-type]
                min_connections=min_connections,
                max_connections=max_connections,
                blocking_timeout=blocking_timeout,
                idle_connection_timeout=idle_connection_timeout,
                hashing_function=hashing_function,
                endpoint_healthcheck_config=endpoint_healthcheck_config,
                **connection_args,
            )

    async def execute_command(self, command: Command[R]) -> None:
        """
        Execute a given memcached command using the connection pool.

        :param command: A memcached command instance to be executed.

        :meta private:
        """
        await self.connection_pool.execute_command(command)

    async def get(self, *keys: KeyT) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        """
        Retrieve one or more items from memcached.

        :param keys: One or more keys identifying the items to be retrieved.
        :return: A dictionary mapping each found key to its corresponding memcached item.
        """
        command = GetCommand[AnyStr](*keys, decode=self.decode_responses, encoding=self.encoding)
        await self.execute_command(command)
        return await command.response

    async def gets(self, *keys: KeyT) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        """
        Retrieve items along with their CAS (Check And Set) identifiers from memcached.

        :param keys: One or more keys identifying the items to be retrieved.
        :return: A dictionary mapping each found key to its corresponding memcached item, including CAS value.
        """
        command = GetsCommand[AnyStr](*keys, decode=self.decode_responses, encoding=self.encoding)
        await self.execute_command(command)
        return await command.response

    async def gat(self, *keys: KeyT, expiry: int) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        """
        Retrieve items from memcached and update their expiration time.

        :param keys: One or more keys identifying the items to be retrieved.
        :param expiry: New expiration time (in seconds) to be applied to the items.
        :return: A dictionary mapping each found key to its corresponding memcached item.
        """
        command = GatCommand[AnyStr](
            *keys, expiry=expiry, decode=self.decode_responses, encoding=self.encoding
        )
        await self.execute_command(command)
        return await command.response

    async def gats(self, *keys: KeyT, expiry: int) -> dict[AnyStr, MemcachedItem[AnyStr]]:
        """
        Retrieve items with CAS identifiers and update their expiration time.

        :param keys: One or more keys identifying the items to be retrieved.
        :param expiry: New expiration time (in seconds) to be applied to the items.
        :return: A dictionary mapping each found key to its corresponding memcached item, including CAS value.
        """
        command = GatsCommand[AnyStr](
            *keys, expiry=expiry, decode=self.decode_responses, encoding=self.encoding
        )
        await self.execute_command(command)
        return await command.response

    @overload
    async def set(
        self, key: KeyT, value: ValueT, /, flags: int = ..., expiry: int = ...
    ) -> bool: ...

    @overload
    async def set(
        self, key: KeyT, value: ValueT, /, flags: int = ..., expiry: int = ..., noreply: bool = ...
    ) -> bool | None: ...

    async def set(
        self, key: KeyT, value: ValueT, /, flags: int = 0, expiry: int = 0, noreply: bool = False
    ) -> bool | None:
        """
        Store a key-value pair in memcached.

        :param key: The key under which the value should be stored.
        :param value: The value to be stored.
        :param flags: Arbitrary flags stored alongside the value (default: 0).
        :param expiry: Expiration time in seconds (default: 0, meaning no expiration).
        :param noreply: If True, the command will not wait for a reply from the server.
        :return: True if the item was stored successfully, False otherwise; returns None if noreply is True.
        """
        command = SetCommand(
            key, value, flags=flags, expiry=expiry, noreply=noreply, encoding=self.encoding
        )
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def cas(
        self,
        key: KeyT,
        value: ValueT,
        cas: int,
        /,
        flags: int = ...,
        expiry: int = ...,
    ) -> bool: ...

    @overload
    async def cas(
        self,
        key: KeyT,
        value: ValueT,
        cas: int,
        /,
        flags: int = ...,
        expiry: int = ...,
        noreply: bool = ...,
    ) -> bool | None: ...

    async def cas(
        self,
        key: KeyT,
        value: ValueT,
        cas: int,
        /,
        flags: int = 0,
        expiry: int = 0,
        noreply: bool = False,
    ) -> bool | None:
        """
        Perform a CAS (Check-And-Set) operation on an item in memcached.

        :param key: The key of the item to update.
        :param value: The new value to store.
        :param cas: The CAS identifier that must match the current CAS value of the item.
        :param flags: Arbitrary flags stored alongside the value (default: 0).
        :param expiry: Expiration time in seconds (default: 0, meaning no expiration).
        :param noreply: If True, the command will not wait for a reply from the server.
        :return: True if the item was updated successfully, False otherwise; returns None if noreply is True.
        """
        command = CheckAndSetCommand(
            key,
            value,
            flags=flags,
            expiry=expiry,
            noreply=noreply,
            cas=cas,
            encoding=self.encoding,
        )
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def add(
        self, key: KeyT, value: ValueT, /, flags: int = ..., expiry: int = ...
    ) -> bool: ...

    @overload
    async def add(
        self, key: KeyT, value: ValueT, /, flags: int = ..., expiry: int = ..., noreply: bool = ...
    ) -> bool | None: ...

    async def add(
        self, key: KeyT, value: ValueT, /, flags: int = 0, expiry: int = 0, noreply: bool = False
    ) -> bool | None:
        """
        Add a key-value pair to memcached only if the key does not already exist.

        :param key: The key under which the value should be added.
        :param value: The value to be stored.
        :param flags: Arbitrary flags stored alongside the value (default: 0).
        :param expiry: Expiration time in seconds (default: 0, meaning no expiration).
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the item was added, False otherwise; returns None if noreply is True.
        """
        command = AddCommand(
            key, value, flags=flags, expiry=expiry, noreply=noreply, encoding=self.encoding
        )
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def append(
        self,
        key: KeyT,
        value: ValueT,
        /,
    ) -> bool: ...

    @overload
    async def append(self, key: KeyT, value: ValueT, /, noreply: bool = ...) -> bool | None: ...

    async def append(self, key: KeyT, value: ValueT, /, noreply: bool = False) -> bool | None:
        """
        Append data to an existing item stored in memcached.

        :param key: The key of the item to append data to.
        :param value: The data to append.
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the append operation succeeded, False otherwise; returns None if noreply is True.
        """
        command = AppendCommand(key, value, noreply=noreply, encoding=self.encoding)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def prepend(self, key: KeyT, value: ValueT, /) -> bool: ...

    @overload
    async def prepend(self, key: KeyT, value: ValueT, /, noreply: bool = ...) -> bool | None: ...

    async def prepend(self, key: KeyT, value: ValueT, /, noreply: bool = False) -> bool | None:
        """
        Prepend data to an existing item stored in memcached.

        :param key: The key of the item to which data should be prepended.
        :param value: The data to prepend.
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the prepend operation succeeded, False otherwise; returns None if noreply is True.
        """
        command = PrependCommand(key, value, noreply=noreply, encoding=self.encoding)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def replace(
        self,
        key: KeyT,
        value: ValueT,
        /,
        flags: int = ...,
        expiry: int = ...,
    ) -> bool: ...

    @overload
    async def replace(
        self,
        key: KeyT,
        value: ValueT,
        /,
        flags: int = ...,
        expiry: int = ...,
        noreply: bool = False,
    ) -> bool | None: ...

    async def replace(
        self, key: KeyT, value: ValueT, /, flags: int = 0, expiry: int = 0, noreply: bool = False
    ) -> bool | None:
        """
        Replace the value for an existing key in memcached.

        :param key: The key whose value is to be replaced.
        :param value: The new value to store.
        :param flags: Arbitrary flags stored alongside the value (default: 0).
        :param expiry: Expiration time in seconds (default: 0, meaning no expiration).
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the replace operation succeeded, False otherwise; returns None if noreply is True.
        """
        command = ReplaceCommand(
            key, value, flags=flags, expiry=expiry, noreply=noreply, encoding=self.encoding
        )
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def incr(self, key: KeyT, value: int, /) -> int | None: ...

    @overload
    async def incr(self, key: KeyT, value: int, /, noreply: bool = ...) -> int | None: ...

    async def incr(self, key: KeyT, value: int, /, noreply: bool = False) -> int | None:
        """
        Increment the numeric value of a key in memcached.

        :param key: The key whose value should be incremented.
        :param value: The amount by which to increment the current value.
        :param noreply: If True, do not wait for a reply from the server.
        :return: The new value after incrementing, or None if noreply is True.
        """
        command = IncrCommand(key, value, noreply)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def decr(self, key: KeyT, value: int, /) -> int | None: ...

    @overload
    async def decr(self, key: KeyT, value: int, /, noreply: bool = ...) -> int | None: ...

    async def decr(self, key: KeyT, value: int, /, noreply: bool = False) -> int | None:
        """
        Decrement the numeric value of a key in memcached.

        :param key: The key whose value should be decremented.
        :param value: The amount by which to decrement the current value.
        :param noreply: If True, do not wait for a reply from the server.
        :return: The new value after decrementing, or None if noreply is True.
        """
        command = DecrCommand(key, value, noreply)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def delete(self, key: KeyT, /) -> bool: ...

    @overload
    async def delete(self, key: KeyT, /, noreply: bool = ...) -> bool | None: ...

    async def delete(self, key: KeyT, /, noreply: bool = False) -> bool | None:
        """
        Delete an item from memcached.

        :param key: The key of the item to be deleted.
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the deletion was successful, False otherwise; returns None if noreply is True.
        """
        command = DeleteCommand(key, noreply=noreply)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    @overload
    async def touch(self, key: KeyT, expiry: int, /) -> bool: ...

    @overload
    async def touch(self, key: KeyT, expiry: int, /, noreply: bool = ...) -> bool | None: ...

    async def touch(self, key: KeyT, expiry: int, /, noreply: bool = False) -> bool | None:
        """
        Update the expiration time for an existing key without modifying its value.

        :param key: The key to update.
        :param expiry: The new expiration time in seconds.
        :param noreply: If True, do not wait for a reply from the server.
        :return: True if the expiration time was updated successfully, False otherwise; returns None if noreply is True.
        """
        command = TouchCommand(key, expiry=expiry, noreply=noreply)
        await self.execute_command(command)
        if noreply:
            return None
        return await command.response

    async def flushall(self, expiry: int = 0, /) -> bool:
        """
        Invalidate all existing items in memcached.

        :param expiry: Delay (in seconds) before flushing all items (default: 0).
        :return: True if the flush operation succeeded, False otherwise.

        .. note:: If the client is configured to use multiple memcached servers
           the result will be True only if all servers succeeded.
        """
        command = FlushAllCommand(expiry=expiry)
        await self.execute_command(command)
        return await command.response

    async def stats(
        self, arg: str | None = None
    ) -> dict[SingleMemcachedInstanceEndpoint, dict[AnyStr, AnyStr]]:
        """
        Retrieve server statistics from memcached.

        :param arg: An optional argument to specify a subset of statistics.
        :return: A mapping of memcached servers to mappings of statistic keys
         and their corresponding values.
        """
        command = StatsCommand[AnyStr](
            arg, decode_responses=self.decode_responses, encoding=self.encoding
        )
        await self.execute_command(command)
        return await command.response

    async def version(self) -> dict[SingleMemcachedInstanceEndpoint, str]:
        """
        Retrieve the memcached server version.

        :return: A mapping of memcached servers to their versions
        """
        command = VersionCommand(noreply=False)
        await self.execute_command(command)
        return await command.response

    def __del__(self) -> None:
        """
        Clean up the client by closing the connection pool.
        """
        if pool := getattr(self, "connection_pool", None):
            pool.close()



from __future__ import annotations

import os
import pathlib
import platform
import sys

import versioneer

__author__ = "Ali-Akber Saifee"
__email__ = "ali@indydevs.org"
__copyright__ = "Copyright 2025, Ali-Akber Saifee"

from setuptools import find_packages, setup

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
PY_IMPLEMENTATION = platform.python_implementation()
USE_MYPYC = False
PURE_PYTHON = os.environ.get("PURE_PYTHON", PY_IMPLEMENTATION != "CPython")


def get_requirements(req_file):
    requirements = []

    for r in open(os.path.join(THIS_DIR, "requirements", req_file)).read().splitlines():
        req = r.strip()

        if req.startswith("-r"):
            requirements.extend(get_requirements(req.replace("-r ", "")))
        elif req:
            requirements.append(req)

    return requirements


_ROOT_DIR = pathlib.Path(__file__).parent

with open(str(_ROOT_DIR / "README.md")) as f:
    long_description = f.read()

if len(sys.argv) > 1 and "--use-mypyc" in sys.argv:
    sys.argv.remove("--use-mypyc")
    USE_MYPYC = True

extensions = []
if not PURE_PYTHON and USE_MYPYC:
    from mypyc.build import mypycify

    extensions += mypycify(
        [],
        debug_level="0",
        strip_asserts=True,
    )
    for ext in extensions:
        if "-Werror" in ext.extra_compile_args:
            ext.extra_compile_args.remove("-Werror")

setup(
    name="memcachio",
    version=versioneer.get_version(),
    description="Python async client for Memcached",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alisaifee/memcachio",
    project_urls={
        "Source": "https://github.com/alisaifee/memcachio",
        "Changes": "https://github.com/alisaifee/memcachio/releases",
        "Documentation": "https://memcachio.readthedocs.org",
    },
    author=__author__,
    author_email=__email__,
    maintainer=__author__,
    maintainer_email=__email__,
    keywords=["Memcached", "asyncio"],
    license="MIT",
    packages=find_packages(exclude=["*tests*"]),
    include_package_data=True,
    package_data={
        "memcachio": ["py.typed"],
    },
    python_requires=">=3.10",
    install_requires=get_requirements("main.txt"),
    cmdclass=versioneer.get_cmdclass(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    ext_modules=extensions,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from .types import SingleMemcachedInstanceEndpoint

if TYPE_CHECKING:
    pass


class MemcachedError(Exception):
    """
    Base exception for any errors raised by the memcached
    servers
    """

    pass


class ClientError(MemcachedError):
    """
    Raised when memcached responds with ``CLIENT_ERROR``
    """

    pass


class ServerError(MemcachedError):
    """
    Raised when memcached responds with ``SERVER_ERROR``
    """

    pass


class NotEnoughData(Exception):
    """
    :meta private:
    """

    def __init__(self, data_read: int):
        self.data_read = data_read
        super().__init__()


class MemcachioConnectionError(ConnectionError):
    """
    Base exception for any connection errors encountered.
    """

    #: The memcached server where the connection error originated from
    endpoint: SingleMemcachedInstanceEndpoint

    def __init__(self, message: str, endpoint: SingleMemcachedInstanceEndpoint):
        self.endpoint = endpoint
        super().__init__(f"{message or 'Connection error'} (memcached server: {endpoint})")


class ConnectionNotAvailable(MemcachioConnectionError):
    """
    Raised when a connection couldn't be acquired from the pool within
    the configured timeout
    """

    def __init__(self, endpoint: SingleMemcachedInstanceEndpoint, timeout: float):
        message = f"Unable to get a connection from the pool in {timeout} seconds"
        super().__init__(message, endpoint=endpoint)


class NoAvailableNodes(ValueError):
    """
    Raised when no nodes are available in the cluster
    """

    pass


class AutoDiscoveryError(MemcachedError):
    """
    Raised when the response for autodiscovery is invalid
    """

    pass



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import AnyStr, Generic, NamedTuple, TypeGuard, cast

#: Acceptable types for a memcached key
KeyT = str | bytes
#: Acceptable types for values to store
ValueT = str | bytes | int


class TCPEndpoint(NamedTuple):
    """
    Location of a memcached server listening on a tcp port
    """

    #: IPV4/6 host address
    host: str
    #: IPV4/6 port
    port: int


#: Path to a memcached server listening on a UDS socket
UnixSocketEndpoint = str | Path


class AWSAutoDiscoveryEndpoint(NamedTuple):
    """
    Location of a memcached auto-discovery endpoint
    for use with AWS Elasticache
    """

    #: IPV4/6 host address
    host: str
    #: IPV4/6 port
    port: int
    #: How often to trigger a refresh to check for updates
    refresh_interval: float


#: The total description of a single memcached instance
SingleMemcachedInstanceEndpoint = UnixSocketEndpoint | TCPEndpoint | tuple[str, int]
#: The total description of either a single memcached instance or a memcached cluster
MemcachedEndpoint = (
    SingleMemcachedInstanceEndpoint
    | Sequence[SingleMemcachedInstanceEndpoint]
    | AWSAutoDiscoveryEndpoint
)


@dataclass
class MemcachedItem(Generic[AnyStr]):
    """
    Data class returned by retrieval commands such as
    :meth:`~memcachio.Client.get`, :meth:`~memcachio.Client.gets`,
    :meth:`~memcachio.Client.gat` and :meth:`~memcachio.Client.gats`
    """

    #: The key of the item
    key: AnyStr
    #: Any flags set on the item
    flags: int
    #: The size (in bytes) of the data stored in the item
    size: int
    #: The CAS value for the item if retrieved
    cas: int | None
    #: The data value of the item
    value: AnyStr


def is_single_server(endpoint: MemcachedEndpoint) -> TypeGuard[SingleMemcachedInstanceEndpoint]:
    if isinstance(endpoint, AWSAutoDiscoveryEndpoint):
        return False
    if isinstance(endpoint, (UnixSocketEndpoint, TCPEndpoint)):
        return True
    if (
        isinstance(endpoint, Sequence)
        and len(endpoint) == 2
        and isinstance(endpoint[0], str)
        and isinstance(endpoint[1], int)
    ):
        return True
    return False


def normalize_single_server_endpoint(
    endpoint: SingleMemcachedInstanceEndpoint,
) -> SingleMemcachedInstanceEndpoint:
    if not isinstance(endpoint, UnixSocketEndpoint):
        return TCPEndpoint(*endpoint)
    return endpoint


def normalize_endpoint(endpoint: MemcachedEndpoint) -> MemcachedEndpoint:
    if is_single_server(endpoint):
        return normalize_single_server_endpoint(endpoint)
    elif isinstance(endpoint, AWSAutoDiscoveryEndpoint):
        return endpoint
    else:
        return [
            normalize_single_server_endpoint(single)
            for single in cast(Sequence[SingleMemcachedInstanceEndpoint], endpoint)
        ]



from __future__ import annotations

from .types import ValueT


def decodedstr(value: ValueT, encoding: str = "utf-8") -> str:
    if isinstance(value, bytes):
        return value.decode(encoding)
    if isinstance(value, str):
        return value
    return str(value)


def bytestr(value: ValueT, encoding: str = "utf-8") -> bytes:
    if isinstance(value, bytes):
        return value
    else:
        return str(value).encode(encoding)
